

class Transaction {

    Map<Integer, TransactionPart> parts = NULL;
    Set<Integer> missingParts = NULL;
    List<Integer> partsPendingSend = NULL;
    bool isComplete = false;
    bool hasLastPart = false;
    Set<KeyValue> keyValueGuardSet = NULL;
    Set<KeyValue> keyValueUpdateSet = NULL;
    bool isDead = false;
    int64_t sequenceNumber = -1;
    int64_t clientLocalSequenceNumber = -1;
    int64_t arbitratorId = -1;
    int64_t machineId = -1;
    Pair<Long, Long> transactionId = NULL;

    int nextPartToSend = 0;
    bool didSendAPartToServer = false;

    TransactionStatus transactionStatus = NULL;

    bool hadServerFailure = false;

    Transaction() {
        parts = new HashMap<Integer, TransactionPart>();
        keyValueGuardSet = new HashSet<KeyValue>();
        keyValueUpdateSet = new HashSet<KeyValue>();
        partsPendingSend = new ArrayList<Integer>();
    }

    void addPartEncode(TransactionPart newPart) {
        parts.put(newPart.getPartNumber(), newPart);
        partsPendingSend.add(newPart.getPartNumber());

        sequenceNumber = newPart.getSequenceNumber();
        arbitratorId = newPart.getArbitratorId();
        transactionId = newPart.getTransactionId();
        clientLocalSequenceNumber = newPart.getClientLocalSequenceNumber();
        machineId = newPart.getMachineId();

        isComplete = true;
    }

    void addPartDecode(TransactionPart newPart) {

        if (isDead) {
            // If dead then just kill this part and move on
            newPart.setDead();
            return;
        }

        sequenceNumber = newPart.getSequenceNumber();
        arbitratorId = newPart.getArbitratorId();
        transactionId = newPart.getTransactionId();
        clientLocalSequenceNumber = newPart.getClientLocalSequenceNumber();
        machineId = newPart.getMachineId();

        TransactionPart previoslySeenPart = parts.put(newPart.getPartNumber(), newPart);

        if (previoslySeenPart != NULL) {
            // Set dead the old one since the new one is a rescued version of this part
            previoslySeenPart.setDead();
        } else if (newPart.isLastPart()) {
            missingParts = new HashSet<Integer>();
            hasLastPart = true;

            for (int i = 0; i < newPart.getPartNumber(); i++) {
                if (parts.get(i) == NULL) {
                    missingParts.add(i);
                }
            }
        }

        if (!isComplete && hasLastPart) {

            // We have seen this part so remove it from the set of missing parts
            missingParts.remove(newPart.getPartNumber());

            // Check if all the parts have been seen
            if (missingParts.size() == 0) {

                // We have all the parts
                isComplete = true;

                // Decode all the parts and create the key value guard and update sets
                decodeTransactionData();
            }
        }
    }

    void addUpdateKV(KeyValue kv) {
        keyValueUpdateSet.add(kv);
    }

    void addGuardKV(KeyValue kv) {
        keyValueGuardSet.add(kv);
    }


    int64_t getSequenceNumber() {
        return sequenceNumber;
    }

    void setSequenceNumber(int64_t _sequenceNumber) {
        sequenceNumber = _sequenceNumber;

        for (Integer i : parts.keySet()) {
            parts.get(i).setSequenceNumber(sequenceNumber);
        }
    }

    int64_t getClientLocalSequenceNumber() {
        return clientLocalSequenceNumber;
    }

    Map<Integer, TransactionPart> getParts() {
        return parts;
    }

    bool didSendAPartToServer() {
        return didSendAPartToServer;
    }

    void resetNextPartToSend() {
        nextPartToSend = 0;
    }

    TransactionPart getNextPartToSend() {
        if ((partsPendingSend.size() == 0) || (partsPendingSend.size() == nextPartToSend)) {
            return NULL;
        }
        TransactionPart part = parts.get(partsPendingSend.get(nextPartToSend));
        nextPartToSend++;
        return part;
    }


    void setServerFailure() {
        hadServerFailure = true;
    }

    bool getServerFailure() {
        return hadServerFailure;
    }


    void resetServerFailure() {
        hadServerFailure = false;
    }


    void setTransactionStatus(TransactionStatus _transactionStatus) {
        transactionStatus = _transactionStatus;
    }

    TransactionStatus getTransactionStatus() {
        return transactionStatus;
    }

    void removeSentParts(List<Integer> sentParts) {
        nextPartToSend = 0;
        if(partsPendingSend.removeAll(sentParts))
        {
            didSendAPartToServer = true;
            transactionStatus.setTransactionSequenceNumber(sequenceNumber);
        }
    }

    bool didSendAllParts() {
        return partsPendingSend.isEmpty();
    }

    Set<KeyValue> getKeyValueUpdateSet() {
        return keyValueUpdateSet;
    }

    int getNumberOfParts() {
        return parts.size();
    }

    int64_t getMachineId() {
        return machineId;
    }

    int64_t getArbitrator() {
        return arbitratorId;
    }

    bool isComplete() {
        return isComplete;
    }

    Pair<Long, Long> getId() {
        return transactionId;
    }

    void setDead() {
        if (isDead) {
            // Already dead
            return;
        }

        // Set dead
        isDead = true;

        // Make all the parts of this transaction dead
        for (Integer partNumber : parts.keySet()) {
            TransactionPart part = parts.get(partNumber);
            part.setDead();
        }
    }

    TransactionPart getPart(int index) {
        return parts.get(index);
    }

    void decodeTransactionData() {

        // Calculate the size of the data section
        int dataSize = 0;
        for (int i = 0; i < parts.keySet().size(); i++) {
            TransactionPart tp = parts.get(i);
            dataSize += tp.getDataSize();
        }

        char[] combinedData = new char[dataSize];
        int currentPosition = 0;

        // Stitch all the data sections together
        for (int i = 0; i < parts.keySet().size(); i++) {
            TransactionPart tp = parts.get(i);
            System.arraycopy(tp.getData(), 0, combinedData, currentPosition, tp.getDataSize());
            currentPosition += tp.getDataSize();
        }

        // Decoder Object
        ByteBuffer bbDecode = ByteBuffer.wrap(combinedData);

        // Decode how many key value pairs need to be decoded
        int numberOfKVGuards = bbDecode.getInt();
        int numberOfKVUpdates = bbDecode.getInt();

        // Decode all the guard key values
        for (int i = 0; i < numberOfKVGuards; i++) {
            KeyValue kv = (KeyValue)KeyValue.decode(bbDecode);
            keyValueGuardSet.add(kv);
        }

        // Decode all the updates key values
        for (int i = 0; i < numberOfKVUpdates; i++) {
            KeyValue kv = (KeyValue)KeyValue.decode(bbDecode);
            keyValueUpdateSet.add(kv);
        }
    }

    bool evaluateGuard(Map<IoTString, KeyValue> committedKeyValueTable, Map<IoTString, KeyValue> speculatedKeyValueTable, Map<IoTString, KeyValue> pendingTransactionSpeculatedKeyValueTable) {
        for (KeyValue kvGuard : keyValueGuardSet) {

            // First check if the key is in the speculative table, this is the value of the latest assumption
            KeyValue kv = NULL;

            // If we have a speculation table then use it first
            if (pendingTransactionSpeculatedKeyValueTable != NULL) {
                kv = pendingTransactionSpeculatedKeyValueTable.get(kvGuard.getKey());
            }

            // If we have a speculation table then use it first
            if ((kv == NULL) && (speculatedKeyValueTable != NULL)) {
                kv = speculatedKeyValueTable.get(kvGuard.getKey());
            }

            if (kv == NULL) {
                // if it is not in the speculative table then check the committed table and use that
                // value as our latest assumption
                kv = committedKeyValueTable.get(kvGuard.getKey());
            }

            if (kvGuard.getValue() != NULL) {
                if ((kv == NULL) || (!kvGuard.getValue().equals(kv.getValue()))) {


                    if (kv != NULL) {
                        System.out.println(kvGuard.getValue() + "       " + kv.getValue());
                    } else {
                        System.out.println(kvGuard.getValue() + "       " + kv);
                    }

                    return false;
                }
            } else {
                if (kv != NULL) {
                    return false;
                }
            }
        }
        return true;
    }
}
